from __future__ import annotations

import logging
import traceback
from enum import Enum
from pathlib import Path
from typing import Any

import click

from exasol.nb_connector.ai_lab_config import StorageBackend
from exasol.nb_connector.cli import reporting as report
from exasol.nb_connector.cli.param_wrappers import ScsSecretOption
from exasol.nb_connector.cli.processing.backend_selector import BackendSelector
from exasol.nb_connector.cli.processing.bucketfs_access import verify_bucketfs_access
from exasol.nb_connector.cli.processing.option_set import (
    SELECT_BACKEND_OPTION,
    USE_ITDE_OPTION,
    OptionSet,
    ScsCliError,
    get_option_set,
    get_scs,
)
from exasol.nb_connector.connections import open_pyexasol_connection
from exasol.nb_connector.secret_store import Secrets

LOG = logging.getLogger(__name__)


def save(
    scs_file: Path,
    backend: StorageBackend,
    use_itde: bool,
    values: dict[str, Any],
):
    """
    Save the provided values to SCS using the keys inferred from backend
    and use_itde.
    """
    scs = get_scs(scs_file)
    config = BackendSelector(scs)
    if not config.matches(backend, use_itde):
        report.warning(
            "Overwriting former SCS configuration for an "
            f"Exasol {config.backend_name} instance."
        )
    options = OptionSet(scs, backend, use_itde)
    values[SELECT_BACKEND_OPTION.arg_name] = backend.name
    values[USE_ITDE_OPTION.arg_name] = use_itde
    values.update(options.default_values(values))
    for arg_name, value in values.items():
        if value is None:
            continue
        option = options.find_option(arg_name)
        if not option.scs_key:
            continue
        if not isinstance(option, ScsSecretOption):
            content = value.value if isinstance(value, Enum) else str(value)
            scs.save(option.scs_key, content)
            continue
        if secret := option.get_secret(interactive=bool(value)):
            scs.save(option.scs_key, secret)


def verify_connection(scs: Secrets) -> None:
    """
    Verify if successful connection to the configured backend is possible.
    Raise an ScsCliError otherwise.
    """
    try:
        open_pyexasol_connection(scs).execute("SELECT 1 FROM DUAL").fetchone()
    except Exception as ex:
        stacktrace = traceback.format_exc()
        raise ScsCliError(
            f"Failed to connect to the configured database {stacktrace}"
        ) from ex
    report.success("Connection to the configured database instance was successful.")


def check_scs(scs_file: Path, connect: bool) -> None:
    """
    Check the SCS content for completeness.  Infer the required keys from
    backend and use_itde if these are contained in the SCS already.

    If parameter `connect` is True then also verify if a connection to the
    configured Exasol database instance is successful.

    The function raises an ScsCliError in any of the following cases:

    * The SCS does not select any backend.

    * The options are incomplete for configuring access to the selected backend.

    * Connecting to the configured backend was requested but failed.

    For Docker-DB (ITDE) the connection is not verified currently, as

    * Startup of ITDE takes quite some time which is considered inconvienient
      for the user

    * Connect does not verify parameters actually configured by the user but
      rather parameters selected by ITDE
    """
    options = get_option_set(scs_file)
    options.check()
    if not connect:
        return
    scs = options.scs
    if BackendSelector(scs).use_itde:
        report.warning(
            "Connection verification for Docker-DB"
            " (via ITDE) is not implemented, yet."
        )
        return
    verify_connection(scs)
    verify_bucketfs_access(scs)


def show_scs_content(scs_file: Path) -> None:
    """
    If the SCS contains a proper backend selection, then show the SCS
    content for this context.
    """
    option_set = get_option_set(scs_file)
    for option in option_set.options:
        value = option.scs_key and option.displayed_value(option_set.scs)
        if value is not None:
            value = value or '""'
            click.echo(f"{option.cli_option()}: {value}")



from exasol.nb_connector.cli.commands.check import check as check
from exasol.nb_connector.cli.commands.configure import configure as configure
from exasol.nb_connector.cli.commands.show import show as show



"""
The UI design aims to provide a convenient way of setting up various configuration parameters.
A user form would consist of one or more blocks of data laid out vertically or horizontally. The positioning of these blocks is defined by the `outer_layout` property.
A block includes a collection of UI elements. Each element is represented by a pair of widgets - a Label with the description of an element and the input widget, e.g. a Text. The layout of the elements within the block is defined by the `group_layout` property.
The layout of the description Label and the input widget is defined by the `row_layout` property. As the name suggests, an input and its description are assumed to be placed next to one another in a row, although such design is not strictly required.
A block may include a header Label where one can put the name of the input data group. `header_style` and `header_layout` define respectively the style and the layout of such labels.
Likewise, the Labels of input elements have their style and layout defined by the `label_style` and `label_layout` properties. The input elements themselves have similar properties - `input_style` and `input_layout`.
A user form may include one or more buttons. Their style and layout are defined as `button_style` and `button_layout`.
"""

SOLID_BORDER: str = "solid 1px"


def config_styles():

    from dataclasses import (
        dataclass,
        field,
    )

    import ipywidgets as widgets

    @dataclass(frozen=True)
    class UiLook:
        header_style: dict = field(default_factory=dict)
        header_layout: widgets.Layout = field(default_factory=widgets.Layout)
        label_style: dict = field(default_factory=dict)
        label_layout: widgets.Layout = field(default_factory=widgets.Layout)
        input_style: dict = field(default_factory=dict)
        input_layout: widgets.Layout = field(default_factory=widgets.Layout)
        button_style: dict = field(default_factory=dict)
        button_layout: widgets.Layout = field(default_factory=widgets.Layout)
        row_layout: widgets.Layout = field(default_factory=widgets.Layout)
        group_layout: widgets.Layout = field(default_factory=widgets.Layout)
        outer_layout: widgets.Layout = field(default_factory=widgets.Layout)

    return UiLook(
        header_style={"background": "Beige"},
        header_layout=widgets.Layout(
            display="flex", justify_content="center", border=SOLID_BORDER
        ),
        label_layout=widgets.Layout(max_width="130px"),
        input_layout=widgets.Layout(max_width="200px"),
        button_style={"button_color": "LightCyan"},
        button_layout=widgets.Layout(border=SOLID_BORDER, margin="12px 0 0 0"),
        row_layout=widgets.Layout(
            display="flex",
            flex_flow="row",
            max_width="350px",
            justify_content="space-between",
            padding="0 0 0 10px",
        ),
        group_layout=widgets.Layout(
            display="flex",
            flex_flow="column",
            border=SOLID_BORDER,
            max_width="350px",
            margin="12px 0 0 0",
        ),
        outer_layout=widgets.Layout(
            display="flex",
            flex_flow="column",
            align_items="stretch",
            padding="0 0 0 50px",
        ),
    )



"""Top-level UI package for notebook-connector.

This package contains the various UI components (access store, common
widgets, configuration UIs, docker-related UIs, etc.).  Subpackages
re-export their primary public entry points so callers can use concise
imports such as::

    from exasol.nb_connector.ui.config import get_generic_config_ui
"""

from . import access  # noqa: F401
from . import common  # noqa: F401
from . import config  # noqa: F401
from . import docker  # noqa: F401

__all__ = ["access", "common", "config", "docker"]



from exasol.nb_connector.slc.package_types import (
    CondaPackageDefinition as CondaPackageDefinition,
)
from exasol.nb_connector.slc.package_types import (
    PipPackageDefinition as PipPackageDefinition,
)
from exasol.nb_connector.slc.script_language_container import (
    ScriptLanguageContainer as ScriptLanguageContainer,
)
from exasol.nb_connector.slc.slc_error import SlcError as SlcError



from pathlib import Path

import ipywidgets as widgets
from IPython import get_ipython

from exasol.nb_connector.secret_store import (
    InvalidPassword,
    Secrets,
)
from exasol.nb_connector.ui.common.popup_message import display_popup
from exasol.nb_connector.ui.common.ui_styles import config_styles

DEFAULT_FILE_NAME = "ai_lab_secure_configuration_storage.sqlite"


def get_scs_location_file_path() -> Path:
    """
    Returns the path to the file where the path of the SCS sqlite database file is stored
    """
    return Path.home() / ".cache" / "notebook-connector" / "scs_file"


def get_sb_store_file():
    try:
        return get_scs_location_file_path().read_text().strip()
    except FileNotFoundError:
        return DEFAULT_FILE_NAME


def set_sb_store_file(value):
    get_scs_location_file_path().parent.mkdir(parents=True, exist_ok=True)
    get_scs_location_file_path().write_text(value)


def get_access_store(root_dir: str = ".") -> widgets.Widget:
    sb_store_file_ = get_sb_store_file()
    ui_look = config_styles()

    header_lbl = widgets.Label(
        value="Configuration Store",
        style=ui_look.header_style,
        layout=ui_look.header_layout,
    )
    file_lbl = widgets.Label(
        value="File Path", style=ui_look.label_style, layout=ui_look.label_layout
    )
    password_lbl = widgets.Label(
        value="Password", style=ui_look.label_style, layout=ui_look.label_layout
    )
    file_txt = widgets.Text(
        value=sb_store_file_, style=ui_look.input_style, layout=ui_look.input_layout
    )
    password_txt = widgets.Password(
        style=ui_look.input_style, layout=ui_look.input_layout
    )
    open_btn = widgets.Button(
        description="Open", style=ui_look.button_style, layout=ui_look.button_layout
    )

    def open_or_create_config_store(btn):
        sb_store_file = file_txt.value
        ipython = get_ipython()
        try:
            ai_lab_config = Secrets(Path(root_dir) / sb_store_file, password_txt.value)
            ai_lab_config.connection()
        except InvalidPassword:
            display_popup(
                "Failed to open the store. Please check that the password is correct"
            )
        else:
            open_btn.icon = "check"
            ipython.push({"ai_lab_config": ai_lab_config}, interactive=True)
        finally:
            set_sb_store_file(sb_store_file)

    def on_value_change(change):
        open_btn.icon = "pen"

    open_btn.on_click(open_or_create_config_store)

    file_txt.observe(on_value_change, names=["value"])
    password_txt.observe(on_value_change, names=["value"])

    group_items = [
        header_lbl,
        widgets.Box([file_lbl, file_txt], layout=ui_look.row_layout),
        widgets.Box([password_lbl, password_txt], layout=ui_look.row_layout),
    ]
    items = [widgets.Box(group_items, layout=ui_look.group_layout), open_btn]
    ui = widgets.Box(items, layout=ui_look.outer_layout)
    return ui
